#define DEBUG

using System;
using System.Text;
using System.Runtime.InteropServices;
using System.Collections;
using System.Collections.Specialized;
using System.Text.RegularExpressions;

namespace browscap {
  public class IniFileName {
    [DllImport("kernel32")]
    static extern int GetPrivateProfileSectionNames(
      [MarshalAs(UnmanagedType.LPArray)] byte[] Result, int Size, string FileName);
    [DllImport("kernel32")]
    static extern int GetPrivateProfileSection(
      string Section, [MarshalAs(UnmanagedType.LPArray)] byte[] Result, int Size, string FileName);
    // inserted for prevent matching of group entries https://github.com/browscap/browscap/issues/3
    [DllImport("kernel32")]
    static extern int GetPrivateProfileString(
      string Section, string Key, string Default, StringBuilder Result, int Size, string FileName);

    private static int CompareByLength(string x, string y)
    {
      if (x == null)
      {
        if (y == null)
        {
          return 0;
        }
        else
        {
          return 1;
        }
      }
      else
      {
        if (y == null)
        {
          return -1;
        }
        else
        {
          int retval = x.Length.CompareTo(y.Length);
          if (retval != 0)
          {
            return -retval;
          }
          else
          {
            return -x.CompareTo(y);
          }
        }
      }
    }

    private string path;
    private string[] sections;

    private string[] GetSectionNames()
    {
      for (int maxsize = 500; true; maxsize *= 2)
      {
        byte[] bytes = new byte[maxsize];
        int size = GetPrivateProfileSectionNames(bytes, maxsize, path);
        if (size < maxsize - 2)
        {
          string Selected = Encoding.ASCII.GetString(bytes, 0, size - (size > 0 ? 1 : 0));
          return Selected.Split(new char[] { '\0' });
        }
      }
    }
    
    public IniFileName(string INIPath) {
#if DEBUG
      int firstTime = Environment.TickCount;
#endif
      path = INIPath;
      sections = GetSectionNames();
      Array.Sort(sections, CompareByLength);
#if DEBUG
      int lastTime = Environment.TickCount;
      Console.WriteLine("load {0} sections in {1} ms", sections.Length, lastTime - firstTime);
#endif
    }
    // inserted for prevent matching of group entries https://github.com/browscap/browscap/issues/3
    public string IniReadValue(string Section, string Key) {
      StringBuilder temp = new StringBuilder(255);
      int i = GetPrivateProfileString(Section, Key, "", temp, 255, path);
      return temp.ToString();
    }
    private string findMatch(string Agent)
    {
      if (sections != null)
      {
        foreach (string SecHead in sections)
        {  // following line changed for https://github.com/browscap/browscap/issues/438
          if ((SecHead.IndexOf("*", 0) == -1) && (SecHead.IndexOf("?", 0) == -1) && (SecHead == Agent))
          {  // inserted for prevent matching of group entries https://github.com/browscap/browscap/issues/3
            if (IniReadValue(SecHead, "parent") != "DefaultProperties")
            {
              return SecHead;
            }
          }
        }
        foreach (string SecHead in sections)
        {
          try
          {  // following line changed for https://github.com/browscap/browscap/issues/438
            if ((SecHead.IndexOf("*", 0) > -1) || (SecHead.IndexOf("?", 0) > -1))
            {
              if (Regex.IsMatch(Agent, "^" + Regex.Escape(SecHead).Replace(@"\*", ".*").Replace(@"\?", ".") + "$"))
              {
#if DEBUG       // following line shows the matching section and RegEx
                Console.WriteLine(SecHead + ": ^" + Regex.Escape(SecHead).Replace(@"\*", ".*").Replace(@"\?", ".") + "$");
#endif
                return SecHead;
              }
            }
          }
          catch (Exception ex)
          {
            Console.WriteLine(ex);
          }
        }
        return "*";
      }
      return "";
    }

    public NameValueCollection getValues(string Agent)
    {
#if DEBUG
      Console.WriteLine("search : [" + Agent + "]");
      int firstTime = Environment.TickCount;
#endif
      string match = findMatch(Agent);
#if DEBUG
      int lastTime = Environment.TickCount;
      Console.WriteLine("find match: [{0}] in {1} ms", match, lastTime - firstTime);
      firstTime = Environment.TickCount;
#endif
      NameValueCollection col = new NameValueCollection();
      do
      {
        string[] entries = new string[0];
        bool goon = true;
        for (int maxsize = 500; goon; maxsize *= 2)
        {
          byte[] bytes = new byte[maxsize];
          int size = GetPrivateProfileSection(match, bytes, maxsize, path);
          if (size < maxsize - 2)
          {
            string section = Encoding.ASCII.GetString(bytes, 0, size - (size > 0 ? 1 : 0));
            entries = section.Split(new char[] { '\0' });
            goon = false;
          }
        }
        match = "";
        if (entries.Length > 0)
        {
          foreach (string entry in entries)
          {
            string[] ent = entry.Split(new char[] { '=' });
            if (ent[0] == "Parent")
            {
              match = ent[1];
            }
            else if (col[ent[0]] == null)
            {
              col.Add(ent[0], ent[1]);
            }
          }
        }
      } while (match != "");
#if DEBUG
      lastTime = Environment.TickCount;
      Console.WriteLine("get values: {0} ms", lastTime - firstTime);
#endif
      return col;
    }
  }
}